pulseeffects (%new_version%-1) unstable; urgency=low

  * Version %new_version% upstream (autobuild)

 -- Mikhail Novosyolov <mikhailnov@dumalogiya.ru>  %day_name%, %day_month% %month_name% %year% %time% %timezone%
